from django.contrib.gis.geos import Point
from rest_framework import pagination, serializers

from rest_framework_gis import serializers as gis_serializers
from rest_framework_gis.fields import GeometrySerializerMethodField

from .models import (
    BoxedLocation,
    GeometryModel,
    LineStringModel,
    Location,
    MultiLineStringModel,
    MultiPointModel,
    MultiPolygonModel,
    PointModel,
    PolygonModel,
)

__all__ = [
    'LocationGeoSerializer',
    'PaginatedLocationGeoSerializer',
    'LocationGeoFeatureSerializer',
    'LocationGeoFeatureSlugSerializer',
    'LocationGeoFeatureFalseIdSerializer',
    'LocationGeoFeatureNoIdSerializer',
    'LocatedFileGeoFeatureSerializer',
    'BoxedLocationGeoFeatureSerializer',
    'LocationGeoFeatureBboxSerializer',
    'LocationGeoFeatureMethodSerializer',
    'NoneGeoFeatureMethodSerializer',
    'PointSerializer',
    'ChildPointSerializer',
    'ListChildPointSerializer',
    'LineStringSerializer',
    'PolygonSerializer',
    'MultiPolygonSerializer',
    'MultiLineStringSerializer',
    'MultiPointSerializer',
    'GeometrySerializerMethodFieldSerializer',
    'GeometrySerializer',
    'BoxedLocationGeoFeatureWithBBoxGeoFieldSerializer',
]


class LocationGeoSerializer(serializers.ModelSerializer):
    """ location geo serializer  """

    details = serializers.HyperlinkedIdentityField(view_name='api_location_details')

    class Meta:
        model = Location
        fields = '__all__'


class PaginatedLocationGeoSerializer(pagination.PageNumberPagination):
    page_size_query_param = 'limit'
    page_size = 40
    max_page_size = 10000


class LocationGeoFeatureSerializer(gis_serializers.GeoFeatureModelSerializer):
    """ location geo serializer  """

    details = serializers.HyperlinkedIdentityField(
        view_name='api_geojson_location_details'
    )
    fancy_name = serializers.SerializerMethodField()

    def get_fancy_name(self, obj):
        return 'Kool %s' % obj.name

    class Meta:
        model = Location
        geo_field = 'geometry'
        fields = '__all__'


class LocationGeoFeatureSlugSerializer(LocationGeoFeatureSerializer):
    """ use slug as id attribute  """

    class Meta:
        model = Location
        geo_field = 'geometry'
        id_field = 'slug'
        fields = ('name', 'slug', 'timestamp')


class LocationGeoFeatureFalseIdSerializer(LocationGeoFeatureSerializer):
    """ id attribute set as False """

    class Meta:
        model = Location
        geo_field = 'geometry'
        id_field = False
        fields = '__all__'


class LocationGeoFeatureNoIdSerializer(LocationGeoFeatureSerializer):
    """
    id attribute left out, issue #82
    see: https://github.com/openwisp/django-rest-framework-gis/issues/82
    """

    class Meta:
        model = Location
        geo_field = 'geometry'
        fields = ('name',)


class LocatedFileGeoFeatureSerializer(gis_serializers.GeoFeatureModelSerializer):
    """ located file geo serializer  """

    details = serializers.HyperlinkedIdentityField(
        view_name='api_geojson_located_file_details'
    )
    fancy_name = serializers.SerializerMethodField()
    file = serializers.FileField(allow_empty_file=True)

    def get_fancy_name(self, obj):
        return 'Nice %s' % obj.name

    class Meta:
        model = Location
        geo_field = 'geometry'
        exclude = []


class BoxedLocationGeoFeatureSerializer(gis_serializers.GeoFeatureModelSerializer):
    """ location geo serializer  """

    details = serializers.HyperlinkedIdentityField(
        view_name='api_geojson_boxedlocation_details'
    )

    class Meta:
        model = BoxedLocation
        geo_field = 'geometry'
        bbox_geo_field = 'bbox_geometry'
        fields = ['name', 'details', 'id', 'timestamp']


class LocationGeoFeatureBboxSerializer(gis_serializers.GeoFeatureModelSerializer):
    class Meta:
        model = Location
        geo_field = 'geometry'
        auto_bbox = True
        exclude = []


class LocationGeoFeatureMethodSerializer(gis_serializers.GeoFeatureModelSerializer):
    new_geometry = gis_serializers.GeometrySerializerMethodField()

    def get_new_geometry(self, obj):
        if obj.name.startswith('hidden'):
            return Point(0.0, 0.0)
        else:
            return obj.geometry

    class Meta:
        model = Location
        geo_field = 'new_geometry'
        exclude = []


class NoneGeoFeatureMethodSerializer(gis_serializers.GeoFeatureModelSerializer):
    new_geometry = gis_serializers.GeometrySerializerMethodField()

    def get_new_geometry(self, obj):
        return None

    class Meta:
        model = Location
        geo_field = 'new_geometry'
        fields = ['name', 'slug', 'id']


class PointSerializer(gis_serializers.GeoFeatureModelSerializer):
    class Meta:
        model = PointModel
        geo_field = 'location'
        fields = '__all__'


class ChildPointSerializer(serializers.Serializer):
    point = PointSerializer()


class ListChildPointSerializer(serializers.Serializer):
    points = PointSerializer(many=True)


class LineStringSerializer(gis_serializers.GeoFeatureModelSerializer):
    class Meta:
        model = LineStringModel
        geo_field = 'points'
        fields = '__all__'


class PolygonSerializer(gis_serializers.GeoFeatureModelSerializer):
    class Meta:
        model = PolygonModel
        geo_field = 'polygon'
        fields = '__all__'


class PolygonModelSerializer(serializers.ModelSerializer):
    class Meta:
        model = PolygonModel
        fields = '__all__'


class MultiPolygonSerializer(gis_serializers.GeoFeatureModelSerializer):
    class Meta:
        model = MultiPolygonModel
        geo_field = 'polygon'
        fields = '__all__'


class MultiLineStringSerializer(gis_serializers.GeoFeatureModelSerializer):
    class Meta:
        model = MultiLineStringModel
        geo_field = 'points'
        fields = '__all__'


class MultiPointSerializer(gis_serializers.GeoFeatureModelSerializer):
    class Meta:
        model = MultiPointModel
        geo_field = 'points'
        fields = '__all__'
        auto_bbox = True


class GeometrySerializer(gis_serializers.GeoFeatureModelSerializer):
    class Meta:
        model = GeometryModel
        geo_field = 'points'
        fields = '__all__'
        auto_bbox = True


class GeometrySerializerMethodFieldSerializer(PointSerializer):
    other_point = GeometrySerializerMethodField()

    @staticmethod
    def get_other_point(obj):
        return Point(obj.location.lat / 2, obj.location.lon / 2)

    class Meta:
        model = Location
        geo_field = 'other_point'
        fields = '__all__'


class BoxedLocationGeoFeatureWithBBoxGeoFieldSerializer(
    BoxedLocationGeoFeatureSerializer
):
    class Meta(BoxedLocationGeoFeatureSerializer.Meta):
        fields = BoxedLocationGeoFeatureSerializer.Meta.fields + [
            BoxedLocationGeoFeatureSerializer.Meta.bbox_geo_field
        ]



from collections import OrderedDict

from rest_framework import pagination
from rest_framework.response import Response


class GeoJsonPagination(pagination.PageNumberPagination):
    """
    A geoJSON implementation of a pagination serializer.
    """

    page_size_query_param = 'page_size'

    def get_paginated_response(self, data):
        return Response(
            OrderedDict(
                [
                    ('type', 'FeatureCollection'),
                    ('count', self.page.paginator.count),
                    ('next', self.get_next_link()),
                    ('previous', self.get_previous_link()),
                    ('features', data['features']),
                ]
            )
        )

    def get_paginated_response_schema(self, view):
        schema = super().get_paginated_response_schema(view)
        schema["properties"]["features"] = schema["properties"].pop("results")
        schema["properties"] = {
            "type": {"type": "string", "enum": ["FeatureCollection"]},
            **schema["properties"],
        }
        return schema



import warnings

from django.contrib.gis.db import models
from rest_framework.schemas.openapi import AutoSchema
from rest_framework.utils import model_meta

from rest_framework_gis.fields import GeometryField, GeometrySerializerMethodField
from rest_framework_gis.serializers import (
    GeoFeatureModelListSerializer,
    GeoFeatureModelSerializer,
)


class GeoFeatureAutoSchema(AutoSchema):
    COORDINATES_SCHEMA_FOR_POINT = {
        "type": "array",
        "items": {"type": "number", "format": "float"},
        "example": [12.9721, 77.5933],
        "minItems": 2,
        "maxItems": 3,
    }

    COORDINATES_SCHEMA_FOR_LINE_STRING = {
        "type": "array",
        "items": COORDINATES_SCHEMA_FOR_POINT,
        "example": [[22.4707, 70.0577], [12.9721, 77.5933]],
        "minItems": 2,
    }

    GEO_FIELD_TO_SCHEMA = {
        models.PointField: {
            "type": {"type": "string", "enum": ["Point"]},
            "coordinates": COORDINATES_SCHEMA_FOR_POINT,
        },
        models.LineStringField: {
            "type": {"type": "string", "enum": ["LineString"]},
            "coordinates": COORDINATES_SCHEMA_FOR_LINE_STRING,
        },
        models.PolygonField: {
            "type": {"type": "string", "enum": ["Polygon"]},
            "coordinates": {
                "type": "array",
                "items": {**COORDINATES_SCHEMA_FOR_LINE_STRING, "minItems": 4},
                "example": [
                    [0.0, 0.0],
                    [0.0, 50.0],
                    [50.0, 50.0],
                    [50.0, 0.0],
                    [0.0, 0.0],
                ],
            },
        },
    }

    GEO_FIELD_TO_SCHEMA[models.GeometryField] = {
        'type': {'type': 'string'},
        'coordinates': {
            'oneOf': [  # If you have custom subclass of GeometryField, Override `oneOf` property.
                GEO_FIELD_TO_SCHEMA[models.PointField],
                GEO_FIELD_TO_SCHEMA[models.LineStringField],
                GEO_FIELD_TO_SCHEMA[models.PolygonField],
            ],
            'example': GEO_FIELD_TO_SCHEMA[models.PolygonField]['coordinates'][
                'example'
            ],
        },
    }

    MULTI_FIELD_MAPPING = {
        models.PointField: models.MultiPointField,
        models.LineStringField: models.MultiLineStringField,
        models.PolygonField: models.MultiPolygonField,
        models.GeometryField: models.GeometryCollectionField,
    }

    for singular_field, multi_field in MULTI_FIELD_MAPPING.items():
        GEO_FIELD_TO_SCHEMA[multi_field] = {
            "type": {"type": "string", "enum": [multi_field.geom_class.__name__]},
            "coordinates": {
                "type": "array",
                "items": GEO_FIELD_TO_SCHEMA[singular_field]["coordinates"],
                "example": [
                    GEO_FIELD_TO_SCHEMA[singular_field]["coordinates"]["example"]
                ],
            },
        }

    def _map_geo_field(self, serializer, geo_field_name):
        field = serializer.fields[geo_field_name]
        if isinstance(field, GeometrySerializerMethodField):
            warnings.warn(
                "Geometry generation for GeometrySerializerMethodField is not supported."
            )
            return {}

        model_field_name = geo_field_name

        geo_field = model_meta.get_field_info(serializer.Meta.model).fields[
            model_field_name
        ]
        try:
            return self.GEO_FIELD_TO_SCHEMA[geo_field.__class__]
        except KeyError:
            warnings.warn(
                "Geometry generation for {field} is not supported.".format(field=field)
            )
            return {}

    def map_field(self, field):
        if isinstance(field, GeoFeatureModelListSerializer):
            return self._map_geo_feature_model_list_serializer(field)

        if isinstance(field, GeometryField):
            return {
                "type": "object",
                "properties": self._map_geo_field(field.parent, field.field_name),
            }

        return super().map_field(field)

    def _map_geo_feature_model_list_serializer(self, serializer):
        return {
            "type": "object",
            "properties": {
                "type": {"type": "string", "enum": ["FeatureCollection"]},
                "features": {
                    "type": "array",
                    "items": self.map_serializer(serializer.child),
                },
            },
        }

    def _map_geo_feature_model_serializer(self, serializer):
        schema = super().map_serializer(serializer)

        geo_json_schema = {
            "type": "object",
            "properties": {"type": {"type": "string", "enum": ["Feature"]}},
        }

        if serializer.Meta.id_field:
            geo_json_schema["properties"]["id"] = schema["properties"].pop(
                serializer.Meta.id_field
            )

        geo_field = serializer.Meta.geo_field
        geo_json_schema["properties"]["geometry"] = {
            "type": "object",
            "properties": self._map_geo_field(serializer, geo_field),
        }
        schema["properties"].pop(geo_field)

        if serializer.Meta.auto_bbox or serializer.Meta.bbox_geo_field:
            geo_json_schema["properties"]["bbox"] = {
                "type": "array",
                "items": {"type": "number"},
                "minItems": 4,
                "maxItems": 4,
                "example": [12.9721, 77.5933, 12.9721, 77.5933],
            }
            if serializer.Meta.bbox_geo_field in schema["properties"]:
                schema["properties"].pop(serializer.Meta.bbox_geo_field)

        geo_json_schema["properties"]["properties"] = schema

        return geo_json_schema

    def map_serializer(self, serializer):
        if isinstance(serializer, GeoFeatureModelListSerializer):
            return self._map_geo_feature_model_list_serializer(serializer)

        if isinstance(serializer, GeoFeatureModelSerializer):
            return self._map_geo_feature_model_serializer(serializer)

        return super().map_serializer(serializer)



"""
unit tests for restframework_gis
"""

import json
import pickle
import sys
from unittest import skipIf

import django
from django.contrib.gis.geos import GEOSGeometry, Point
from django.test import TestCase

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

import rest_framework
from django.core.exceptions import ImproperlyConfigured

from rest_framework_gis import serializers as gis_serializers
from rest_framework_gis.fields import GeoJsonDict

from .models import LocatedFile, Location, Nullable
from .serializers import LocationGeoSerializer

is_pre_drf_39 = not rest_framework.VERSION.startswith('3.9')


class TestRestFrameworkGis(TestCase):
    def setUp(self):
        self.location_list_url = reverse('api_location_list')
        self.geojson_location_list_url = reverse('api_geojson_location_list')
        self.geos_error_message = (
            'Invalid format: string or unicode input unrecognized as GeoJSON,'
            ' WKT EWKT or HEXEWKB.'
        )
        self.gdal_error_message = (
            'Unable to convert to python object:'
            ' Invalid geometry pointer returned from "OGR_G_CreateGeometryFromJson".'
        )
        if django.VERSION >= (2, 0, 0):
            self.value_error_message = (
                "Unable to convert to python object:"
                " String input unrecognized as WKT EWKT, and HEXEWKB."
            )
        else:
            self.value_error_message = (
                "Unable to convert to python object:"
                " String or unicode input unrecognized as WKT EWKT, and HEXEWKB."
            )
        self.type_error_message = (
            "Unable to convert to python object: Improper geometry input type:"
        )

    def _create_locations(self):
        self.l1 = Location.objects.create(
            id=1,
            name='l1',
            slug='l1',
            geometry='POINT (13.0078125000020002 42.4234565179379999)',
        )
        self.l2 = Location.objects.create(
            id=2,
            name='l2',
            slug='l2',
            geometry='POINT (12.0078125000020002 43.4234565179379999)',
        )

    def test_get_location_list(self):
        response = self.client.get(self.location_list_url)
        self.assertEqual(response.status_code, 200)

    def test_post_location_list_geojson(self):
        self.assertEqual(Location.objects.count(), 0)
        data = {
            "name": "geojson input test",
            "geometry": {
                "type": "GeometryCollection",
                "geometries": [
                    {"type": "Point", "coordinates": [12.492324113849, 41.890307434153]}
                ],
            },
        }
        response = self.client.post(
            self.location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)
        data = {
            "name": "geojson input test2",
            "geometry": {
                "type": "Point",
                "coordinates": [12.492324113849, 41.890307434153],
            },
        }
        response = self.client.post(
            self.location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 2)

    def test_post_location_list_geojson_as_multipartformdata(self):
        """ emulate sending geojson string in webform """
        self.assertEqual(Location.objects.count(), 0)
        data = {
            "name": "geojson input test",
            "geometry": json.dumps(
                {
                    "type": "GeometryCollection",
                    "geometries": [
                        {
                            "type": "Point",
                            "coordinates": [12.492324113849, 41.890307434153],
                        }
                    ],
                }
            ),
        }
        response = self.client.post(self.location_list_url, data)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)

    def test_post_HTML_browsable_api(self):
        self.assertEqual(Location.objects.count(), 0)
        data = {
            "name": "geojson input test2",
            "slug": "prova",
            "geometry": json.dumps(
                {
                    "type": "GeometryCollection",
                    "geometries": [
                        {
                            "type": "Point",
                            "coordinates": [12.492324113849, 41.890307434153],
                        }
                    ],
                }
            ),
        }
        response = self.client.post(
            self.location_list_url, data, HTTP_ACCEPT='text/html'
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)
        location = Location.objects.all()[0]
        self.assertEqual(location.name, 'geojson input test2')
        self.assertEqual(location.slug, 'prova')

    def test_post_location_list_WKT(self):
        self.assertEqual(Location.objects.count(), 0)
        data = {
            'name': 'WKT input test',
            'geometry': 'POINT (12.492324113849 41.890307434153)',
        }
        response = self.client.post(self.location_list_url, data)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)

    def test_post_location_list_EWKT(self):
        self.assertEqual(Location.objects.count(), 0)
        data = {
            'name': 'EWKT input test',
            'geometry': 'SRID=28992;POINT(221160 600204)',
        }
        response = self.client.post(self.location_list_url, data)
        expected_coords = (6.381495826183805, 53.384066927384985)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)
        for lat, lon in zip(
            Location.objects.get(name='EWKT input test').geometry.coords,
            expected_coords,
        ):
            self.assertAlmostEqual(lat, lon, places=5)

    def test_post_location_list_WKT_as_json(self):
        self.assertEqual(Location.objects.count(), 0)
        data = {
            'name': 'WKT input test',
            'geometry': 'POINT (12.492324113849 41.890307434153)',
        }
        response = self.client.post(
            self.location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)

    def test_post_location_list_empty_geometry(self):
        data = {'name': 'empty input test'}
        response = self.client.post(self.location_list_url, data)
        self.assertEqual(response.data['geometry'][0], 'This field is required.')
        data = {'name': 'empty input test', 'geometry': ''}
        response = self.client.post(self.location_list_url, data)
        self.assertEqual(response.data['geometry'][0], 'This field is required.')
        data = {'name': 'empty input test'}
        response = self.client.post(
            self.location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.data['geometry'][0], 'This field is required.')
        data = {'name': 'empty input test', 'geometry': ''}
        response = self.client.post(
            self.location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.data['geometry'][0], 'This field is required.')

    def test_post_location_list_invalid_WKT(self):
        data = {'name': 'WKT wrong input test', 'geometry': 'I AM OBVIOUSLY WRONG'}
        response = self.client.post(
            self.location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(Location.objects.count(), 0)
        self.assertEqual(response.data['geometry'][0], self.value_error_message)
        # repeat as multipart form data
        response = self.client.post(self.location_list_url, data)
        self.assertEqual(response.data['geometry'][0], self.value_error_message)
        data = {
            'name': 'I AM MODERATELY WRONG',
            'geometry': 'POINT (12.492324113849, 41.890307434153)',
        }
        response = self.client.post(
            self.location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.data['geometry'][0], self.geos_error_message)
        # repeat as multipart form data
        response = self.client.post(self.location_list_url, data)
        self.assertEqual(response.data['geometry'][0], self.geos_error_message)

    def test_post_location_list_invalid_geojson(self):
        data = {
            "name": "quite wrong",
            "geometry": {
                "type": "ARRRR",
                "dasdas": [{"STtype": "PTUAMAoint", "NNAare": "rgon"}],
            },
        }
        response = self.client.post(
            self.location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.data['geometry'][0], self.gdal_error_message)
        data = {"name": "very wrong", "geometry": ['a', 'b', 'c']}
        response = self.client.post(
            self.location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.data['geometry'][0][0:65], self.type_error_message)
        data = {"name": "very wrong", "geometry": False}
        response = self.client.post(
            self.location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.data['geometry'][0][0:65], self.type_error_message)
        data = {"name": "very wrong", "geometry": {"value": {"nested": ["yo"]}}}
        response = self.client.post(
            self.location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.data['geometry'][0], self.gdal_error_message)

    def test_geojson_format(self):
        """ test geojson format """
        location = Location.objects.create(
            name='geojson test', geometry='POINT (135.0 45.0)'
        )
        url = reverse('api_geojson_location_details', args=[location.id])
        expected = {
            'id': location.id,
            'type': 'Feature',
            'properties': {
                'details': "http://testserver/geojson/%s/" % location.id,
                'name': 'geojson test',
                'fancy_name': 'Kool geojson test',
                'timestamp': None,
                'slug': 'geojson-test',
            },
            'geometry': {'type': 'Point', 'coordinates': [135.0, 45.0]},
        }
        response = self.client.get(url)
        if sys.version_info > (3, 0, 0):
            self.assertCountEqual(json.dumps(response.data), json.dumps(expected))
        else:
            self.assertItemsEqual(json.dumps(response.data), json.dumps(expected))
        response = self.client.get(url, HTTP_ACCEPT='text/html')
        self.assertContains(response, "Kool geojson test")

    def test_geojson_id_attribute(self):
        location = Location.objects.create(
            name='geojson test', geometry='POINT (10.1 10.1)'
        )
        url = reverse('api_geojson_location_details', args=[location.id])
        response = self.client.get(url)
        self.assertEqual(response.data['id'], location.id)

    def test_geojson_id_attribute_slug(self):
        location = Location.objects.create(
            name='geojson test', geometry='POINT (10.1 10.1)'
        )
        url = reverse('api_geojson_location_slug_details', args=[location.slug])
        response = self.client.get(url)
        self.assertEqual(response.data['id'], location.slug)

    def test_geojson_false_id_attribute_slug(self):
        location = Location.objects.create(
            name='falseid test', geometry='POINT (10.1 10.1)'
        )
        url = reverse('api_geojson_location_falseid_details', args=[location.id])
        response = self.client.get(url)
        self.assertEqual(response.data['properties']['name'], 'falseid test')
        with self.assertRaises(KeyError):
            response.data['id']

    def test_geojson_no_id_attribute_slug(self):
        location = Location.objects.create(
            name='noid test', geometry='POINT (10.1 10.1)'
        )
        url = reverse('api_geojson_location_noid_details', args=[location.id])
        response = self.client.get(url)
        self.assertEqual(response.data['properties']['name'], 'noid test')
        with self.assertRaises(KeyError):
            response.data['id']

    def test_geojson_filefield_attribute(self):
        located_file = LocatedFile.objects.create(
            name='geojson filefield test', geometry='POINT (10.1 10.1)'
        )
        url = reverse('api_geojson_located_file_details', args=[located_file.id])
        response = self.client.get(url)
        self.assertEqual(response.data['properties']['file'], None)

    def test_post_geojson_location_list(self):
        self.assertEqual(Location.objects.count(), 0)
        data = {
            "type": "Feature",
            "properties": {"name": "point?", "details": "ignore this"},
            "geometry": {"type": "Point", "coordinates": [10.1, 10.1]},
        }
        response = self.client.post(
            self.geojson_location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)
        url = reverse(
            'api_geojson_location_details',
            args=[Location.objects.order_by('-id')[0].id],
        )
        response = self.client.get(url)
        self.assertEqual(response.data['properties']['name'], "point?")
        self.assertEqual(response.data['geometry']['type'], "Point")
        self.assertEqual(
            json.dumps(response.data['geometry']['coordinates']), "[10.1, 10.1]"
        )
        self.assertNotEqual(response.data['properties']['details'], "ignore this")

    def test_post_geojson_location_list_HTML(self):
        self.assertEqual(Location.objects.count(), 0)
        data = {
            "type": "Feature",
            "properties": {"name": "point?", "details": "ignore this"},
            "geometry": {"type": "Point", "coordinates": [10.1, 10.1]},
        }
        response = self.client.post(
            self.geojson_location_list_url,
            data=json.dumps(data),
            content_type='application/json',
            HTTP_ACCEPT='text/html',
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)
        url = reverse(
            'api_geojson_location_details',
            args=[Location.objects.order_by('-id')[0].id],
        )
        response = self.client.get(url)
        self.assertEqual(response.data['properties']['name'], "point?")
        self.assertEqual(response.data['geometry']['type'], "Point")
        self.assertEqual(
            json.dumps(response.data['geometry']['coordinates']), "[10.1, 10.1]"
        )
        self.assertNotEqual(response.data['properties']['details'], "ignore this")

    def test_post_invalid_geojson_location_list(self):
        data = {
            "type": "Feature",
            "properties": {"details": "ignore this"},
            "geometry": {"type": "Point", "coordinates": [10.1, 10.1]},
        }
        response = self.client.post(
            self.geojson_location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(Location.objects.count(), 0)
        self.assertEqual(response.data['name'][0], "This field is required.")
        data = {
            "type": "Feature",
            "properties": {"name": "point?"},
            "geometry": {"type": "Point", "WRONG": {}},
        }
        response = self.client.post(
            self.geojson_location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(Location.objects.count(), 0)
        self.assertEqual(response.data['geometry'][0], self.gdal_error_message)

    def test_post_geojson_location_list_WKT(self):
        self.assertEqual(Location.objects.count(), 0)
        data = {
            "type": "Feature",
            "properties": {"name": "point?"},
            "geometry": "POINT (10.1 10.1)",
        }
        response = self.client.post(
            self.geojson_location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)
        url = reverse(
            'api_geojson_location_details',
            args=[Location.objects.order_by('-id')[0].id],
        )
        response = self.client.get(url)
        self.assertEqual(response.data['properties']['name'], "point?")
        self.assertEqual(response.data['geometry']['type'], "Point")
        self.assertEqual(
            json.dumps(response.data['geometry']['coordinates']), "[10.1, 10.1]"
        )

    def test_geofeatured_model_serializer_compatible_with_geomodel_serializer(self):
        self.assertEqual(Location.objects.count(), 0)
        data = {
            "name": "geojson input test",
            "geometry": {
                "type": "GeometryCollection",
                "geometries": [
                    {"type": "Point", "coordinates": [12.492324113849, 41.890307434153]}
                ],
            },
        }
        response = self.client.post(
            self.geojson_location_list_url,
            data=json.dumps(data),
            content_type='application/json',
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)

    def test_geofeatured_model_post_as_multipartformdata(self):
        """ emulate sending geojson string in webform """
        self.assertEqual(Location.objects.count(), 0)
        data = {
            "name": "geojson input test",
            "geometry": json.dumps(
                {"type": "Point", "coordinates": [12.492324113849, 41.890307434153]}
            ),
        }
        response = self.client.post(self.location_list_url, data)
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)
        self.assertEqual(response.data['geometry']['type'], "Point")

    def test_HTML_browsable_geojson_location_list(self):
        response = self.client.get(
            self.geojson_location_list_url, HTTP_ACCEPT='text/html'
        )
        self.assertEqual(response.status_code, 200)
        self._create_locations()
        response = self.client.get(
            self.geojson_location_list_url, HTTP_ACCEPT='text/html'
        )
        self.assertContains(response, 'l1')
        self.assertContains(response, 'l2')

    def test_post_geojson_location_list_HTML_web_form(self):
        self.assertEqual(Location.objects.count(), 0)
        data = {
            "name": "HTML test",
            "geometry": json.dumps({"type": "Point", "coordinates": [10.1, 10.1]}),
        }
        response = self.client.post(
            self.geojson_location_list_url, data, HTTP_ACCEPT='text/html'
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)
        location = Location.objects.all()[0]
        self.assertEqual(location.name, "HTML test")
        self.assertEqual(location.geometry.geom_type, "Point")

    def test_post_geojson_location_list_HTML_web_form_WKT(self):
        self.assertEqual(Location.objects.count(), 0)
        data = {"name": "HTML test WKT", "geometry": "POINT (10.1 10.1)"}
        response = self.client.post(
            self.geojson_location_list_url, data, HTTP_ACCEPT='text/html'
        )
        self.assertEqual(response.status_code, 201)
        self.assertEqual(Location.objects.count(), 1)
        location = Location.objects.all()[0]
        self.assertEqual(location.name, "HTML test WKT")

    @skipIf(is_pre_drf_39, 'Skip this test if DRF < 3.9')
    def test_geojson_HTML_widget_value(self):
        self._create_locations()
        response = self.client.get(
            self.geojson_location_list_url, HTTP_ACCEPT='text/html'
        )
        self.assertContains(response, '<textarea name="geometry"')
        self.assertContains(response, '"type": "Point"')
        self.assertContains(response, '"coordinates": [')

    @skipIf(not is_pre_drf_39, 'Skip this test if DRF >= 3.9')
    def test_geojson_HTML_widget_value_pre_drf_39(self):
        self._create_locations()
        response = self.client.get(
            self.geojson_location_list_url, HTTP_ACCEPT='text/html'
        )
        self.assertContains(response, '<textarea name="geometry"')
        self.assertContains(response, '&quot;type&quot;: &quot;Point&quot;')
        self.assertContains(response, '&quot;coordinates&quot;: [')
        # TODO: remove this when python 2 will be deprecated
        self.assertNotContains(
            response, 'u&#39;type&#39;: u&#39;Point&#39;, u&#39;coordinates&#39;:'
        )

    def test_patch_geojson_location(self):
        location = Location.objects.create(
            name='geojson patch test', geometry='POINT (135.0 45.0)'
        )
        url = reverse('api_geojson_location_details', args=[location.id])
        data = {
            "properties": {"name": "geojson successful patch test"},
            "geometry": {"type": "Point", "coordinates": [10.1, 10.1]},
        }
        response = self.client.generic(
            'PATCH', url, json.dumps(data), content_type='application/json'
        )
        self.assertEqual(response.status_code, 200)
        location_reloaded = Location.objects.get(pk=location.id)
        self.assertEqual(location_reloaded.name, 'geojson successful patch test')
        self.assertEqual(
            location_reloaded.geometry, Point(10.1, 10.1, srid=location.geometry.srid)
        )

    def test_patch_geojson_location_wo_changing_geometry(self):
        location = Location.objects.create(
            name='geojson patch test', geometry='POINT (135.0 45.0)'
        )
        url = reverse('api_geojson_location_details', args=[location.id])
        data = {"properties": {"name": "geojson successful patch test"}}
        response = self.client.generic(
            'PATCH', url, json.dumps(data), content_type='application/json'
        )
        self.assertEqual(response.status_code, 200)
        location_reloaded = Location.objects.get(pk=location.id)
        self.assertEqual(location_reloaded.name, 'geojson successful patch test')
        self.assertEqual(
            location_reloaded.geometry, Point(135.0, 45.0, srid=location.geometry.srid)
        )

    def test_geometry_serializer_method_field(self):
        location = Location.objects.create(
            name='geometry serializer method test', geometry='POINT (135.0 45.0)'
        )
        location_loaded = Location.objects.get(pk=location.id)
        self.assertEqual(location_loaded.name, 'geometry serializer method test')
        self.assertEqual(
            location_loaded.geometry, Point(135.0, 45.0, srid=location.geometry.srid)
        )
        url = reverse('api_geojson_location_details_hidden', args=[location.id])
        data = {"properties": {"name": "hidden geometry"}}
        response = self.client.generic(
            'PATCH', url, json.dumps(data), content_type='application/json'
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['properties']['name'], 'hidden geometry')
        self.assertEqual(response.data['geometry']['type'], 'Point')
        self.assertEqual(response.data['geometry']['coordinates'], [0.0, 0.0])

    def test_geometry_serializer_method_field_none(self):
        location = Location.objects.create(
            name='None value', geometry='POINT (135.0 45.0)'
        )
        location_loaded = Location.objects.get(pk=location.id)
        self.assertEqual(
            location_loaded.geometry, Point(135.0, 45.0, srid=location.geometry.srid)
        )
        url = reverse('api_geojson_location_details_none', args=[location.id])
        response = self.client.generic('GET', url, content_type='application/json')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.data['properties']['name'], 'None value')
        self.assertEqual(response.data['geometry'], None)

    def test_nullable_empty_geometry(self):
        empty = Nullable(name='empty', geometry='POINT EMPTY')
        empty.full_clean()
        empty.save()
        url = reverse('api_geojson_nullable_details', args=[empty.id])
        response = self.client.generic('GET', url, content_type='application/json')
        self.assertIsNotNone(response.data['geometry'])
        self.assertEqual(response.data['geometry']['coordinates'], [])

    def test_nullable_null_geometry(self):
        empty = Nullable(name='empty', geometry=None)
        empty.full_clean()
        empty.save()
        url = reverse('api_geojson_nullable_details', args=[empty.id])
        response = self.client.generic('GET', url, content_type='application/json')
        self.assertIsNone(response.data['geometry'])

    def test_geometry_field_to_representation_none(self):
        self._create_locations()
        f = LocationGeoSerializer(instance=self.l1).fields['geometry']
        self.assertIsNone(f.to_representation(None))

    def test_geometry_empty_representation(self):
        self._create_locations()
        f = LocationGeoSerializer(instance=self.l1).fields['geometry']
        geom_types = (
            'POINT',
            'LINESTRING',
            'LINEARRING',
            'POLYGON',
            'MULTIPOINT',
            'MULTILINESTRING',
            'MULTIPOLYGON',
        )
        for geom_type in geom_types:
            with self.subTest(geom_type=geom_type):
                value = f.to_representation(GEOSGeometry('{} EMPTY'.format(geom_type)))
                self.assertIsNotNone(value)
                if geom_type == 'LINEARRING':
                    geom_type = 'LINESTRING'
                self.assertEqual(value['type'].upper(), geom_type)
                self.assertEqual(value['coordinates'], [])
        # GEOMETRYCOLLECTION needs different handling
        value = f.to_representation(GEOSGeometry('GEOMETRYCOLLECTION EMPTY'))
        self.assertIsNotNone(value)
        self.assertEqual(value['type'].upper(), 'GEOMETRYCOLLECTION')
        self.assertEqual(value['geometries'], [])

    def test_no_geo_field_improperly_configured(self):
        class LocationGeoFeatureSerializer(gis_serializers.GeoFeatureModelSerializer):
            class Meta:
                model = Location

        with self.assertRaises(ImproperlyConfigured):
            LocationGeoFeatureSerializer()

    def test_exclude_geo_field_improperly_configured(self):
        self._create_locations()

        class LocationGeoFeatureSerializer(gis_serializers.GeoFeatureModelSerializer):
            class Meta:
                model = Location
                geo_field = 'geometry'
                exclude = ('geometry',)

        with self.assertRaises(ImproperlyConfigured):
            LocationGeoFeatureSerializer(instance=self.l1)

    def test_geojson_pagination(self):
        self._create_locations()
        response = self.client.get(self.geojson_location_list_url)
        self.assertEqual(response.data['type'], 'FeatureCollection')
        self.assertEqual(len(response.data['features']), 2)
        response = self.client.get(
            '{0}?page_size=1'.format(self.geojson_location_list_url)
        )
        self.assertEqual(response.data['type'], 'FeatureCollection')
        self.assertEqual(len(response.data['features']), 1)
        self.assertIn('next', response.data)
        self.assertIn('previous', response.data)

    def test_pickle(self):
        geometry = GEOSGeometry('POINT (30 10)')
        geojsondict = GeoJsonDict(
            (('type', geometry.geom_type), ('coordinates', geometry.coords),)
        )
        pickled = pickle.dumps(geojsondict)
        restored = pickle.loads(pickled)
        self.assertEqual(restored, geojsondict)

    def test_geometrycollection_geojson(self):
        """ test geometry collection geojson behaviour """
        location = Location.objects.create(
            name='geometry collection geojson test',
            geometry='GEOMETRYCOLLECTION ('
            'POINT (135.0 45.0),'
            'LINESTRING (135.0 45.0,140.0 50.0,145.0 55.0),'
            'POLYGON ((135.0 45.0,140.0 50.0,145.0 55.0,135.0 45.0)))',
        )
        url = reverse('api_geojson_location_details', args=[location.id])
        expected = {
            'id': location.id,
            'type': 'Feature',
            'properties': {
                'details': "http://testserver/geojson/%s/" % location.id,
                'name': 'geometry collection geojson test',
                'fancy_name': 'Kool geometry collection geojson test',
                'timestamp': None,
                'slug': 'geometry-collection-geojson-test',
            },
            'geometry': {
                'type': 'GeometryCollection',
                'geometries': [
                    {'type': 'Point', 'coordinates': [135.0, 45.0]},
                    {
                        'type': 'LineString',
                        'coordinates': [[135.0, 45.0], [140.0, 50.0], [145.0, 55.0]],
                    },
                    {
                        'type': 'Polygon',
                        'coordinates': [
                            [
                                [135.0, 45.0],
                                [140.0, 50.0],
                                [145.0, 55.0],
                                [135.0, 45.0],
                            ]
                        ],
                    },
                ],
            },
        }
        response = self.client.get(url)
        if sys.version_info > (3, 0, 0):
            self.assertCountEqual(json.dumps(response.data), json.dumps(expected))
        else:
            self.assertItemsEqual(json.dumps(response.data), json.dumps(expected))
        self.assertContains(response, "Kool geometry collection geojson test")



from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import generics

from rest_framework_gis.filters import (
    DistanceToPointFilter,
    DistanceToPointOrderingFilter,
    GeoFilterSet,
    GeometryFilter,
    InBBoxFilter,
    TMSTileFilter,
)
from rest_framework_gis.pagination import GeoJsonPagination

from .models import BoxedLocation, LocatedFile, Location, Nullable, PolygonModel
from .serializers import (
    BoxedLocationGeoFeatureSerializer,
    LocatedFileGeoFeatureSerializer,
    LocationGeoFeatureBboxSerializer,
    LocationGeoFeatureFalseIdSerializer,
    LocationGeoFeatureMethodSerializer,
    LocationGeoFeatureNoIdSerializer,
    LocationGeoFeatureSerializer,
    LocationGeoFeatureSlugSerializer,
    LocationGeoSerializer,
    NoneGeoFeatureMethodSerializer,
    PaginatedLocationGeoSerializer,
    PolygonModelSerializer,
)


class LocationList(generics.ListCreateAPIView):
    model = Location
    serializer_class = LocationGeoSerializer
    queryset = Location.objects.all()
    pagination_class = PaginatedLocationGeoSerializer


location_list = LocationList.as_view()


class LocationDetails(generics.RetrieveUpdateDestroyAPIView):
    model = Location
    serializer_class = LocationGeoSerializer
    queryset = Location.objects.all()


location_details = LocationDetails.as_view()


class GeojsonLocationList(generics.ListCreateAPIView):
    model = Location
    serializer_class = LocationGeoFeatureSerializer
    queryset = Location.objects.all()
    pagination_class = GeoJsonPagination


geojson_location_list = GeojsonLocationList.as_view()


class GeojsonLocationContainedInBBoxList(generics.ListAPIView):
    model = Location
    serializer_class = LocationGeoFeatureSerializer
    queryset = Location.objects.all()
    bbox_filter_field = 'geometry'
    filter_backends = (InBBoxFilter,)


geojson_location_contained_in_bbox_list = GeojsonLocationContainedInBBoxList.as_view()


class GeojsonLocationOverlapsBBoxList(GeojsonLocationContainedInBBoxList):
    bbox_filter_include_overlapping = True


geojson_location_overlaps_bbox_list = GeojsonLocationOverlapsBBoxList.as_view()


class GeojsonLocationContainedInTileList(generics.ListAPIView):
    model = Location
    serializer_class = LocationGeoFeatureSerializer
    queryset = Location.objects.all()
    bbox_filter_field = 'geometry'
    filter_backends = (TMSTileFilter,)


geojson_location_contained_in_tile_list = GeojsonLocationContainedInTileList.as_view()


class GeojsonLocationOverlapsTileList(GeojsonLocationContainedInTileList):
    bbox_filter_include_overlapping = True


geojson_location_overlaps_tile_list = GeojsonLocationOverlapsTileList.as_view()


class GeojsonLocationWithinDistanceOfPointList(generics.ListAPIView):
    model = Location
    serializer_class = LocationGeoFeatureSerializer
    distance_filter_convert_meters = True
    queryset = Location.objects.all()
    distance_filter_field = 'geometry'
    filter_backends = (DistanceToPointFilter,)


geojson_location_within_distance_of_point_list = (
    GeojsonLocationWithinDistanceOfPointList.as_view()
)


class GeojsonLocationWithinDegreesOfPointList(GeojsonLocationWithinDistanceOfPointList):
    distance_filter_convert_meters = False  # Default setting


geojson_location_within_degrees_of_point_list = (
    GeojsonLocationWithinDegreesOfPointList.as_view()
)


class GeojsonLocationOrderDistanceToPointList(generics.ListAPIView):
    model = Location
    serializer_class = LocationGeoFeatureSerializer
    queryset = Location.objects.all()
    distance_ordering_filter_field = 'geometry'
    filter_backends = (DistanceToPointOrderingFilter,)


geojson_location_order_distance_to_point_list = (
    GeojsonLocationOrderDistanceToPointList.as_view()
)


class GeojsonLocationDetails(generics.RetrieveUpdateDestroyAPIView):
    model = Location
    serializer_class = LocationGeoFeatureSerializer
    queryset = Location.objects.all()


geojson_location_details = GeojsonLocationDetails.as_view()


class GeojsonLocationDetailsHidden(generics.RetrieveUpdateDestroyAPIView):
    model = Location
    serializer_class = LocationGeoFeatureMethodSerializer
    queryset = Location.objects.all()


geojson_location_details_hidden = GeojsonLocationDetailsHidden.as_view()


class GeojsonLocationDetailsNone(generics.RetrieveUpdateDestroyAPIView):
    model = Location
    serializer_class = NoneGeoFeatureMethodSerializer
    queryset = Location.objects.all()


geojson_location_details_none = GeojsonLocationDetailsNone.as_view()


class GeojsonLocationSlugDetails(generics.RetrieveUpdateDestroyAPIView):
    model = Location
    lookup_field = 'slug'
    serializer_class = LocationGeoFeatureSlugSerializer
    queryset = Location.objects.all()


geojson_location_slug_details = GeojsonLocationSlugDetails.as_view()


class GeojsonLocationFalseIdDetails(generics.RetrieveUpdateDestroyAPIView):
    model = Location
    serializer_class = LocationGeoFeatureFalseIdSerializer
    queryset = Location.objects.all()


geojson_location_falseid_details = GeojsonLocationFalseIdDetails.as_view()


class GeojsonLocationNoIdDetails(generics.RetrieveUpdateDestroyAPIView):
    model = Location
    serializer_class = LocationGeoFeatureNoIdSerializer
    queryset = Location.objects.all()


geojson_location_noid_details = GeojsonLocationNoIdDetails.as_view()


class LocationFilter(GeoFilterSet):
    contains_properly = GeometryFilter(
        field_name='geometry', lookup_expr='contains_properly'
    )

    class Meta:
        model = Location
        fields = ['contains_properly']


class GeojsonLocationContainedInGeometry(generics.ListAPIView):
    queryset = Location.objects.all()
    serializer_class = LocationGeoSerializer
    filter_class = LocationFilter

    filter_backends = (DjangoFilterBackend,)


geojson_contained_in_geometry = GeojsonLocationContainedInGeometry.as_view()


class GeojsonLocatedFileDetails(generics.RetrieveUpdateDestroyAPIView):
    model = LocatedFile
    serializer_class = LocatedFileGeoFeatureSerializer
    queryset = LocatedFile.objects.all()


geojson_located_file_details = GeojsonLocatedFileDetails.as_view()


class GeojsonBoxedLocationDetails(generics.RetrieveUpdateDestroyAPIView):
    model = BoxedLocation
    serializer_class = BoxedLocationGeoFeatureSerializer
    queryset = BoxedLocation.objects.all()


geojson_boxedlocation_details = GeojsonBoxedLocationDetails.as_view()


class GeojsonBoxedLocationList(generics.ListCreateAPIView):
    model = BoxedLocation
    serializer_class = BoxedLocationGeoFeatureSerializer
    queryset = BoxedLocation.objects.all()


geojson_boxedlocation_list = GeojsonBoxedLocationList.as_view()


class GeojsonLocationBboxList(generics.ListCreateAPIView):
    model = Location
    serializer_class = LocationGeoFeatureBboxSerializer
    queryset = Location.objects.all()


geojson_location_bbox_list = GeojsonLocationBboxList.as_view()


class GeojsonNullableDetails(generics.RetrieveUpdateDestroyAPIView):
    model = Location
    serializer_class = LocationGeoFeatureSerializer
    queryset = Nullable.objects.all()


geojson_nullable_details = GeojsonNullableDetails.as_view()


class ModelViewWithPolygon(generics.RetrieveUpdateDestroyAPIView):
    model = PolygonModel
    serializer_class = PolygonModelSerializer



from django.contrib.gis.db import models
from django.utils.text import slugify

__all__ = [
    'Location',
    'LocatedFile',
    'BoxedLocation',
    'GeometryModel',
    'Nullable',
    'PointModel',
    'LineStringModel',
    'PolygonModel',
    'MultiPolygonModel',
    'MultiLineStringModel',
    'MultiPointModel',
]


class BaseModel(models.Model):
    name = models.CharField(max_length=32)
    slug = models.SlugField(max_length=128, unique=True, blank=True)
    timestamp = models.DateTimeField(null=True, blank=True)
    geometry = models.GeometryField()

    class Meta:
        abstract = True

    def __str__(self):
        return self.name

    def _generate_slug(self):
        if self.slug == '' or self.slug is None:
            try:
                name = str(self.name)
            except NameError:
                name = self.name
            self.slug = slugify(name)

    def clean(self):
        self._generate_slug()

    def save(self, *args, **kwargs):
        self._generate_slug()
        super().save(*args, **kwargs)


class Location(BaseModel):
    pass


class LocatedFile(BaseModel):
    file = models.FileField(upload_to='located_files', blank=True, null=True)


class BoxedLocation(BaseModel):
    bbox_geometry = models.PolygonField()


class Nullable(BaseModel):
    geometry = models.GeometryField(blank=True, null=True)


class SchemaBaseModel(models.Model):
    random_field1 = models.CharField(max_length=32)
    random_field2 = models.IntegerField()

    class Meta:
        abstract = True


class PointModel(SchemaBaseModel):
    location = models.PointField()


class LineStringModel(SchemaBaseModel):
    points = models.LineStringField()


class PolygonModel(SchemaBaseModel):
    polygon = models.PolygonField()


class GeometryModel(SchemaBaseModel):
    points = models.GeometryField()


class MultiPointModel(SchemaBaseModel):
    points = models.MultiPointField()


class MultiLineStringModel(SchemaBaseModel):
    points = models.MultiLineStringField()


class MultiPolygonModel(SchemaBaseModel):
    polygon = models.MultiPolygonField()



from django.urls import path

from . import views

urlpatterns = [
    path('', views.location_list, name='api_location_list'),
    path('<int:pk>', views.location_details, name='api_location_details'),
    # geojson
    path('geojson/', views.geojson_location_list, name='api_geojson_location_list'),
    path(
        'geojson/<int:pk>/',
        views.geojson_location_details,
        name='api_geojson_location_details',
    ),
    path(
        'geojson-nullable/<int:pk>/',
        views.geojson_nullable_details,
        name='api_geojson_nullable_details',
    ),
    path(
        'geojson_hidden/<int:pk>/',
        views.geojson_location_details_hidden,
        name='api_geojson_location_details_hidden',
    ),
    path(
        'geojson_none/<int:pk>/',
        views.geojson_location_details_none,
        name='api_geojson_location_details_none',
    ),
    path(
        'geojson/<slug:slug>/',
        views.geojson_location_slug_details,
        name='api_geojson_location_slug_details',
    ),
    path(
        'geojson-falseid/<int:pk>/',
        views.geojson_location_falseid_details,
        name='api_geojson_location_falseid_details',
    ),
    path(
        'geojson-noid/<int:pk>/',
        views.geojson_location_noid_details,
        name='api_geojson_location_noid_details',
    ),
    # file
    path(
        'geojson-file/<int:pk>/',
        views.geojson_located_file_details,
        name='api_geojson_located_file_details',
    ),
    # geojson with bbox with its own geometry field
    path(
        'geojson-with-bbox/',
        views.geojson_boxedlocation_list,
        name='api_geojson_boxedlocation_list',
    ),
    path(
        'geojson-with-bbox/<int:pk>/',
        views.geojson_boxedlocation_details,
        name='api_geojson_boxedlocation_details',
    ),
    # geojson with bbox with autogenerated bbox
    path(
        'geojson-with-auto-bbox/',
        views.geojson_location_bbox_list,
        name='api_geojson_location_bbox_list',
    ),
    # Filters
    path(
        'filters/contained_in_bbox',
        views.geojson_location_contained_in_bbox_list,
        name='api_geojson_location_list_contained_in_bbox_filter',
    ),
    path(
        'filters/overlaps_bbox',
        views.geojson_location_overlaps_bbox_list,
        name='api_geojson_location_list_overlaps_bbox_filter',
    ),
    path(
        'filters/contained_in_geometry',
        views.geojson_contained_in_geometry,
        name='api_geojson_contained_in_geometry',
    ),
    path(
        'filters/contained_in_tile',
        views.geojson_location_contained_in_tile_list,
        name='api_geojson_location_list_contained_in_tile_filter',
    ),
    path(
        'filters/overlaps_tile',
        views.geojson_location_overlaps_tile_list,
        name='api_geojson_location_list_overlaps_tile_filter',
    ),
    path(
        'filters/within_distance_of_point',
        views.geojson_location_within_distance_of_point_list,
        name='api_geojson_location_list_within_distance_of_point_filter',
    ),
    path(
        'filters/within_degrees_of_point',
        views.geojson_location_within_degrees_of_point_list,
        name='api_geojson_location_list_within_degrees_of_point_filter',
    ),
    path(
        'filters/order_distance_to_point',
        views.geojson_location_order_distance_to_point_list,
        name='api_geojson_location_order_distance_to_point_list_filter',
    ),
]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.4 on 2016-04-01 17:00
from __future__ import unicode_literals

import django.contrib.gis.db.models.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='BoxedLocation',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('name', models.CharField(max_length=32)),
                ('slug', models.SlugField(blank=True, max_length=128, unique=True)),
                ('timestamp', models.DateTimeField(blank=True, null=True)),
                (
                    'geometry',
                    django.contrib.gis.db.models.fields.GeometryField(srid=4326),
                ),
                (
                    'bbox_geometry',
                    django.contrib.gis.db.models.fields.PolygonField(srid=4326),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='LocatedFile',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('name', models.CharField(max_length=32)),
                ('slug', models.SlugField(blank=True, max_length=128, unique=True)),
                ('timestamp', models.DateTimeField(blank=True, null=True)),
                (
                    'geometry',
                    django.contrib.gis.db.models.fields.GeometryField(srid=4326),
                ),
                (
                    'file',
                    models.FileField(blank=True, null=True, upload_to='located_files'),
                ),
            ],
            options={'abstract': False},
        ),
        migrations.CreateModel(
            name='Location',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('name', models.CharField(max_length=32)),
                ('slug', models.SlugField(blank=True, max_length=128, unique=True)),
                ('timestamp', models.DateTimeField(blank=True, null=True)),
                (
                    'geometry',
                    django.contrib.gis.db.models.fields.GeometryField(srid=4326),
                ),
            ],
            options={'abstract': False},
        ),
    ]



# Generated by Django 2.2.9 on 2019-12-30 12:24

import django.contrib.gis.db.models.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_restframework_gis_tests', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Nullable',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('name', models.CharField(max_length=32)),
                ('slug', models.SlugField(blank=True, max_length=128, unique=True)),
                ('timestamp', models.DateTimeField(blank=True, null=True)),
                (
                    'geometry',
                    django.contrib.gis.db.models.fields.GeometryField(
                        blank=True, null=True, srid=4326
                    ),
                ),
            ],
            options={'abstract': False},
        ),
    ]



import json
from collections import OrderedDict

from django.contrib.gis.gdal import GDALException
from django.contrib.gis.geos import GEOSException, GEOSGeometry
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from rest_framework.fields import Field, SerializerMethodField

__all__ = ['GeometryField', 'GeometrySerializerMethodField']


class GeometryField(Field):
    """
    A field to handle GeoDjango Geometry fields
    """

    type_name = 'GeometryField'

    def __init__(self, precision=None, remove_duplicates=False, **kwargs):
        self.precision = precision
        self.remove_dupes = remove_duplicates
        super().__init__(**kwargs)
        self.style.setdefault('base_template', 'textarea.html')

    def to_representation(self, value):
        if isinstance(value, dict) or value is None:
            return value
        # we expect value to be a GEOSGeometry instance
        if value.geojson:
            geojson = GeoJsonDict(value.geojson)
        # in this case we're dealing with an empty point
        else:
            geojson = GeoJsonDict({'type': value.geom_type, 'coordinates': []})
        if geojson['type'] == 'GeometryCollection':
            geometries = geojson.get('geometries')
        else:
            geometries = [geojson]
        for geometry in geometries:
            if self.precision is not None:
                geometry['coordinates'] = self._recursive_round(
                    geometry['coordinates'], self.precision
                )
            if self.remove_dupes:
                geometry['coordinates'] = self._rm_redundant_points(
                    geometry['coordinates'], geometry['type']
                )
        return geojson

    def to_internal_value(self, value):
        if value == '' or value is None:
            return value
        if isinstance(value, GEOSGeometry):
            # value already has the correct representation
            return value
        if isinstance(value, dict):
            value = json.dumps(value)
        try:
            return GEOSGeometry(value)
        except (GEOSException):
            raise ValidationError(
                _(
                    'Invalid format: string or unicode input unrecognized as GeoJSON, WKT EWKT or HEXEWKB.'
                )
            )
        except (ValueError, TypeError, GDALException) as e:
            raise ValidationError(
                _('Unable to convert to python object: {}'.format(str(e)))
            )

    def validate_empty_values(self, data):
        if data == '':
            self.fail('required')
        return super().validate_empty_values(data)

    def _recursive_round(self, value, precision):
        """
        Round all numbers within an array or nested arrays
            value: number or nested array of numbers
            precision: integer valueue of number of decimals to keep
        """
        if hasattr(value, '__iter__'):
            return tuple(self._recursive_round(v, precision) for v in value)
        return round(value, precision)

    def _rm_redundant_points(self, geometry, geo_type):
        """
        Remove redundant coordinate pairs from geometry
            geometry: array of coordinates or nested-array of coordinates
            geo_type: GeoJSON type attribute for provided geometry, used to
                     determine structure of provided `geometry` argument
        """
        if geo_type in ('MultiPoint', 'LineString'):
            close = geo_type == 'LineString'
            output = []
            for coord in geometry:
                coord = tuple(coord)
                if not output or coord != output[-1]:
                    output.append(coord)
            if close and len(output) == 1:
                output.append(output[0])
            return tuple(output)
        if geo_type in ('MultiLineString', 'Polygon'):
            return [self._rm_redundant_points(c, 'LineString') for c in geometry]
        if geo_type == 'MultiPolygon':
            return [self._rm_redundant_points(c, 'Polygon') for c in geometry]
        return geometry


class GeometrySerializerMethodField(SerializerMethodField):
    def to_representation(self, value):
        value = super().to_representation(value)
        if value is not None:
            # we expect value to be a GEOSGeometry instance
            return GeoJsonDict(value.geojson)
        else:
            return None


class GeoJsonDict(OrderedDict):
    """
    Used for serializing GIS values to GeoJSON values
    TODO: remove this when support for python 2.6/2.7 will be dropped
    """

    def __init__(self, *args, **kwargs):
        """
        If a string is passed attempt to pass it through json.loads,
        because it should be a geojson formatted string.
        """
        if args and isinstance(args[0], str):
            try:
                geojson = json.loads(args[0])
                args = (geojson,)
            except ValueError:
                pass
        super().__init__(*args, **kwargs)

    def __str__(self):
        """
        Avoid displaying strings like
        ``{ 'type': u'Point', 'coordinates': [12, 32] }``
        in DRF browsable UI inputs (python 2.6/2.7)
        see: https://github.com/openwisp/django-rest-framework-gis/pull/60
        """
        return json.dumps(self)



from unittest import SkipTest

import rest_framework
from django.test import RequestFactory, TestCase
from packaging import version
from rest_framework.generics import RetrieveAPIView
from rest_framework.request import Request
from rest_framework.schemas.openapi import SchemaGenerator

from rest_framework_gis.pagination import GeoJsonPagination
from rest_framework_gis.schema import GeoFeatureAutoSchema

from .serializers import (
    BoxedLocationGeoFeatureWithBBoxGeoFieldSerializer,
    ChildPointSerializer,
    GeometrySerializer,
    GeometrySerializerMethodFieldSerializer,
    LineStringSerializer,
    ListChildPointSerializer,
    MultiLineStringSerializer,
    MultiPointSerializer,
    MultiPolygonSerializer,
    PointSerializer,
)
from .views import (
    GeojsonBoxedLocationDetails,
    GeojsonLocationContainedInBBoxList,
    GeojsonLocationContainedInTileList,
    GeojsonLocationWithinDistanceOfPointList,
    ModelViewWithPolygon,
    geojson_location_list,
)

is_pre_drf_312 = version.parse(rest_framework.VERSION) < version.parse('3.12')

if is_pre_drf_312:
    raise SkipTest('Skip this test if DRF < 3.12')


def create_request(path):
    factory = RequestFactory()
    request = Request(factory.get(path))
    return request


def create_view(view_cls, method, request):
    generator = SchemaGenerator()
    view = generator.create_view(view_cls.as_view(), method, request)
    return view


class TestSchemaGeneration(TestCase):
    def test_point_field_outer_most_gis_serializer(self):
        class TestPointFieldView(RetrieveAPIView):
            serializer_class = PointSerializer

        path = "/"
        method = "GET"

        view = create_view(TestPointFieldView, "POST", create_request("/"))
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        content = inspector.map_serializer(serializer)
        content.pop("type", None)
        content["properties"]["properties"].pop("type", None)
        self.assertEqual(
            content,
            {
                "properties": {
                    "type": {"type": "string", "enum": ["Feature"]},
                    "id": {"type": "integer", "readOnly": True},
                    "geometry": {
                        "type": "object",
                        "properties": {
                            "type": {"type": "string", "enum": ["Point"]},
                            "coordinates": {
                                "type": "array",
                                "items": {"type": "number", "format": "float"},
                                "example": [12.9721, 77.5933],
                                "minItems": 2,
                                "maxItems": 3,
                            },
                        },
                    },
                    "properties": {
                        "properties": {
                            "random_field1": {"type": "string", "maxLength": 32},
                            "random_field2": {
                                "type": "integer",
                                "maximum": 2147483647,
                                "minimum": -2147483648,
                            },
                        },
                        "required": ["random_field1", "random_field2", "location"],
                    },
                }
            },
        )

    def test_point_field_inner_geo_serializer(self):
        class TestPointFieldView(RetrieveAPIView):
            serializer_class = ChildPointSerializer

        path = "/"
        method = "GET"

        view = create_view(TestPointFieldView, "POST", create_request("/"))
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        content = inspector.map_serializer(serializer)
        content.pop("type", None)
        content["properties"]["point"].pop("type", None)
        content["properties"]["point"]["properties"]["properties"].pop("type", None)
        self.assertEqual(
            content,
            {
                "properties": {
                    "point": {
                        "properties": {
                            "type": {"type": "string", "enum": ["Feature"]},
                            "id": {"type": "integer", "readOnly": True},
                            "geometry": {
                                "type": "object",
                                "properties": {
                                    "type": {"type": "string", "enum": ["Point"]},
                                    "coordinates": {
                                        "type": "array",
                                        "items": {"type": "number", "format": "float"},
                                        "example": [12.9721, 77.5933],
                                        "minItems": 2,
                                        "maxItems": 3,
                                    },
                                },
                            },
                            "properties": {
                                "properties": {
                                    "random_field1": {
                                        "type": "string",
                                        "maxLength": 32,
                                    },
                                    "random_field2": {
                                        "type": "integer",
                                        "maximum": 2147483647,
                                        "minimum": -2147483648,
                                    },
                                },
                                "required": [
                                    "random_field1",
                                    "random_field2",
                                    "location",
                                ],
                            },
                        }
                    }
                },
                "required": ["point"],
            },
        )

    def test_point_field_inner_geo_list_serializer(self):
        class TestPointFieldView(RetrieveAPIView):
            serializer_class = ListChildPointSerializer

        path = "/"
        method = "GET"

        view = create_view(TestPointFieldView, "POST", create_request("/"))
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        content = inspector.map_serializer(serializer)
        content.pop("type", None)
        content["properties"]["points"].pop("type", None)
        content["properties"]["points"]["properties"]["features"]["items"][
            "properties"
        ]["properties"].pop("type", None)
        self.assertEqual(
            content,
            {
                "properties": {
                    "points": {
                        "properties": {
                            "type": {"type": "string", "enum": ["FeatureCollection"]},
                            "features": {
                                "type": "array",
                                "items": {
                                    "type": "object",
                                    "properties": {
                                        "type": {"type": "string", "enum": ["Feature"]},
                                        "id": {"type": "integer", "readOnly": True},
                                        "geometry": {
                                            "type": "object",
                                            "properties": {
                                                "type": {
                                                    "type": "string",
                                                    "enum": ["Point"],
                                                },
                                                "coordinates": {
                                                    "type": "array",
                                                    "items": {
                                                        "type": "number",
                                                        "format": "float",
                                                    },
                                                    "example": [12.9721, 77.5933],
                                                    "minItems": 2,
                                                    "maxItems": 3,
                                                },
                                            },
                                        },
                                        "properties": {
                                            "properties": {
                                                "random_field1": {
                                                    "type": "string",
                                                    "maxLength": 32,
                                                },
                                                "random_field2": {
                                                    "type": "integer",
                                                    "maximum": 2147483647,
                                                    "minimum": -2147483648,
                                                },
                                            },
                                            "required": [
                                                "random_field1",
                                                "random_field2",
                                                "location",
                                            ],
                                        },
                                    },
                                },
                            },
                        }
                    }
                },
                "required": ["points"],
            },
        )

    def test_line_string_field(self):
        class TestLineStringFieldView(RetrieveAPIView):
            serializer_class = LineStringSerializer

        path = "/"
        method = "GET"

        view = create_view(TestLineStringFieldView, "POST", create_request("/"))
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        content = inspector.map_serializer(serializer)
        geometry_schema = content["properties"]["geometry"]
        geometry_schema.pop("type", None)
        self.assertEqual(
            geometry_schema,
            {
                "properties": {
                    "type": {"type": "string", "enum": ["LineString"]},
                    "coordinates": {
                        "type": "array",
                        "items": {
                            "type": "array",
                            "items": {"type": "number", "format": "float"},
                            "example": [12.9721, 77.5933],
                            "minItems": 2,
                            "maxItems": 3,
                        },
                        "example": [[22.4707, 70.0577], [12.9721, 77.5933]],
                        "minItems": 2,
                    },
                }
            },
        )

    def test_multi_polygon(self):
        class TestMultiPolygonFieldView(RetrieveAPIView):
            serializer_class = MultiPolygonSerializer

        path = "/"
        method = "GET"

        view = create_view(TestMultiPolygonFieldView, "POST", create_request("/"))
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        content = inspector.map_serializer(serializer)
        geometry_schema = content["properties"]["geometry"]
        geometry_schema.pop("type", None)
        self.assertEqual(
            geometry_schema,
            {
                "properties": {
                    "type": {"type": "string", "enum": ["MultiPolygon"]},
                    "coordinates": {
                        "type": "array",
                        "items": {
                            "type": "array",
                            "items": {
                                "type": "array",
                                "items": {
                                    "type": "array",
                                    "items": {"type": "number", "format": "float"},
                                    "example": [12.9721, 77.5933],
                                    "minItems": 2,
                                    "maxItems": 3,
                                },
                                "example": [[22.4707, 70.0577], [12.9721, 77.5933]],
                                "minItems": 4,
                            },
                            "example": [
                                [0.0, 0.0],
                                [0.0, 50.0],
                                [50.0, 50.0],
                                [50.0, 0.0],
                                [0.0, 0.0],
                            ],
                        },
                        "example": [
                            [
                                [0.0, 0.0],
                                [0.0, 50.0],
                                [50.0, 50.0],
                                [50.0, 0.0],
                                [0.0, 0.0],
                            ]
                        ],
                    },
                }
            },
        )

    def test_multi_line_string_field(self):
        class TestMultiLineStringFieldView(RetrieveAPIView):
            serializer_class = MultiLineStringSerializer

        path = "/"
        method = "GET"

        view = create_view(TestMultiLineStringFieldView, "POST", create_request("/"))
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        content = inspector.map_serializer(serializer)
        geometry_schema = content["properties"]["geometry"]
        geometry_schema.pop("type", None)
        self.assertEqual(
            geometry_schema,
            {
                "properties": {
                    "type": {"type": "string", "enum": ["MultiLineString"]},
                    "coordinates": {
                        "type": "array",
                        "items": {
                            "type": "array",
                            "items": {
                                "type": "array",
                                "items": {"type": "number", "format": "float"},
                                "example": [12.9721, 77.5933],
                                "minItems": 2,
                                "maxItems": 3,
                            },
                            "example": [[22.4707, 70.0577], [12.9721, 77.5933]],
                            "minItems": 2,
                        },
                        "example": [[[22.4707, 70.0577], [12.9721, 77.5933]]],
                    },
                }
            },
        )

    def test_multi_point(self):
        class TestMultiLineStringFieldView(RetrieveAPIView):
            serializer_class = MultiPointSerializer

        path = "/"
        method = "GET"

        view = create_view(TestMultiLineStringFieldView, "POST", create_request("/"))
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        content = inspector.map_serializer(serializer)
        geometry_schema = content["properties"]["geometry"]
        geometry_schema.pop("type", None)
        self.assertEqual(
            geometry_schema,
            {
                "properties": {
                    "type": {"type": "string", "enum": ["MultiPoint"]},
                    "coordinates": {
                        "type": "array",
                        "items": {
                            "type": "array",
                            "items": {"type": "number", "format": "float"},
                            "example": [12.9721, 77.5933],
                            "minItems": 2,
                            "maxItems": 3,
                        },
                        "example": [[12.9721, 77.5933]],
                    },
                }
            },
        )

    def test_geometry_field(self):
        class GeometrySerializerFieldView(RetrieveAPIView):
            serializer_class = GeometrySerializer

        path = "/"
        method = "GET"

        view = create_view(GeometrySerializerFieldView, "POST", create_request("/"))
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        content = inspector.map_serializer(serializer)
        geometry_schema = content["properties"]["geometry"]
        geometry_schema.pop("type", None)
        geometry_schema['properties'].pop("coordinates", None)
        self.assertEqual(geometry_schema, {"properties": {"type": {"type": "string"}}})

    def check_bbox_schema(self):
        class TestMultiLineStringFieldView(RetrieveAPIView):
            serializer_class = MultiPointSerializer

        path = "/"
        method = "GET"

        view = create_view(TestMultiLineStringFieldView, "POST", create_request("/"))
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        content = inspector.map_serializer(serializer)
        bbox_schema = content["properties"]["bbox"]
        self.assertEqual(
            bbox_schema,
            {
                "type": "array",
                "items": {"type": "number"},
                "minItems": 4,
                "maxItems": 4,
                "example": [12.9721, 77.5933, 12.9721, 77.5933],
            },
        )

    def test_warning_for_geometry_serializer_method_field(self):
        class TestGeometrySerializerMethodField(RetrieveAPIView):
            serializer_class = GeometrySerializerMethodFieldSerializer

        path = "/"
        method = "GET"

        view = create_view(
            TestGeometrySerializerMethodField, "POST", create_request("/")
        )
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        with self.assertWarns(Warning):
            inspector.map_serializer(serializer)

    def test_schema_for_bbox_geo_field(self):
        path = "/"
        method = "GET"

        class GeojsonBoxedLocationDetailsWithBBoxGeoFieldView(
            GeojsonBoxedLocationDetails
        ):
            serializer_class = BoxedLocationGeoFeatureWithBBoxGeoFieldSerializer

        view = create_view(
            GeojsonBoxedLocationDetailsWithBBoxGeoFieldView, "GET", create_request("/")
        )
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        content = inspector.map_serializer(serializer)
        bbox_schema = content["properties"]["bbox"]
        self.assertNotIn(
            "bbox_geometry", content["properties"]["properties"]["properties"]
        )
        self.assertEqual(
            bbox_schema,
            {
                "type": "array",
                "items": {"type": "number"},
                "minItems": 4,
                "maxItems": 4,
                "example": [12.9721, 77.5933, 12.9721, 77.5933],
            },
        )


class TestPaginationSchemaGeneration(TestCase):
    def test_geo_json_pagination_schema(self):
        generated_schema = GeoJsonPagination().get_paginated_response_schema(
            geojson_location_list
        )
        self.assertIn("features", generated_schema["properties"])
        generated_schema["properties"].pop("features")
        self.assertDictEqual(
            generated_schema,
            {
                "type": "object",
                "properties": {
                    "type": {"type": "string", "enum": ["FeatureCollection"]},
                    "count": {"type": "integer", "example": 123},
                    "next": {
                        "type": "string",
                        "nullable": True,
                        "format": "uri",
                        "example": "http://api.example.org/accounts/?page=4",
                    },
                    "previous": {
                        "type": "string",
                        "nullable": True,
                        "format": "uri",
                        "example": "http://api.example.org/accounts/?page=2",
                    },
                },
            },
        )


class TestRestFrameworkGisFiltersSchema(TestCase):
    def test_in_BBox_filter_schema(self):
        path = "/"
        method = "GET"
        view = create_view(
            GeojsonLocationContainedInBBoxList, "GET", create_request("/")
        )
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        generated_schema = inspector.get_filter_parameters(path, method)
        self.assertDictEqual(
            generated_schema[0],
            {
                "name": "in_bbox",
                "required": False,
                "in": "query",
                "description": "Specify a bounding box as filter: in_bbox=min_lon,min_lat,max_lon,max_lat",
                "schema": {
                    "type": "array",
                    "items": {"type": "float"},
                    "minItems": 4,
                    "maxItems": 4,
                    "example": [0, 0, 10, 10],
                },
                "style": "form",
                "explode": False,
            },
        )

    def test_in_TMS_filter_schema(self):
        path = "/"
        method = "GET"
        view = create_view(
            GeojsonLocationContainedInTileList, "GET", create_request("/")
        )
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        generated_schema = inspector.get_filter_parameters(path, method)
        self.assertDictEqual(
            generated_schema[0],
            {
                "name": "tile",
                "required": False,
                "in": "query",
                "description": "Specify a bounding box filter defined by a TMS tile address: tile=Z/X/Y",
                "schema": {"type": "string", "example": "12/56/34"},
            },
        )

    def test_distance_to_point_filter(self):
        path = "/"
        method = "GET"
        view = create_view(
            GeojsonLocationWithinDistanceOfPointList, "GET", create_request("/")
        )
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        generated_schema = inspector.get_filter_parameters(path, method)
        self.assertListEqual(
            generated_schema,
            [
                {
                    "name": "dist",
                    "required": False,
                    "in": "query",
                    "schema": {"type": "number", "format": "float", "default": 1000},
                    "description": "Represents **Distance** in **Distance to point** filter. "
                    "Default value is used only if ***point*** is passed.",
                },
                {
                    "name": "point",
                    "required": False,
                    "in": "query",
                    "description": "Point represented in **x,y** format. "
                    "Represents **point** in **Distance to point filter**",
                    "schema": {
                        "type": "array",
                        "items": {"type": "float"},
                        "minItems": 2,
                        "maxItems": 2,
                        "example": [0, 10],
                    },
                    "style": "form",
                    "explode": False,
                },
            ],
        )

    def test_geometry_field(self):
        path = "/"
        method = "GET"
        view = create_view(ModelViewWithPolygon, "GET", create_request("/"))
        inspector = GeoFeatureAutoSchema()
        inspector.view = view
        serializer = inspector.get_serializer(path, method)
        content = inspector.map_serializer(serializer)

        self.assertEqual(
            content,
            {
                "type": "object",
                "properties": {
                    "id": {'type': 'integer', 'readOnly': True},
                    'polygon': {
                        'properties': {
                            'coordinates': {
                                'example': [
                                    [0.0, 0.0],
                                    [0.0, 50.0],
                                    [50.0, 50.0],
                                    [50.0, 0.0],
                                    [0.0, 0.0],
                                ],
                                'items': {
                                    'example': [[22.4707, 70.0577], [12.9721, 77.5933]],
                                    'items': {
                                        'example': [12.9721, 77.5933],
                                        'items': {'format': 'float', 'type': 'number'},
                                        'maxItems': 3,
                                        'minItems': 2,
                                        'type': 'array',
                                    },
                                    'minItems': 4,
                                    'type': 'array',
                                },
                                'type': 'array',
                            },
                            'type': {'enum': ['Polygon'], 'type': 'string'},
                        },
                        'type': 'object',
                    },
                    "random_field1": {"type": "string", "maxLength": 32},
                    "random_field2": {
                        "type": "integer",
                        "maximum": 2147483647,
                        "minimum": -2147483648,
                    },
                },
                "required": ["random_field1", "random_field2", "polygon"],
            },
        )



# Generated by Django 3.0.4 on 2020-04-03 06:52

import django.contrib.gis.db.models.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_restframework_gis_tests", "0002_nullable"),
    ]

    operations = [
        migrations.CreateModel(
            name="LineStringModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("random_field1", models.CharField(max_length=32)),
                ("random_field2", models.IntegerField()),
                (
                    "points",
                    django.contrib.gis.db.models.fields.LineStringField(srid=4326),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="MultiLineStringModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("random_field1", models.CharField(max_length=32)),
                ("random_field2", models.IntegerField()),
                (
                    "points",
                    django.contrib.gis.db.models.fields.MultiLineStringField(srid=4326),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="MultiPointModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("random_field1", models.CharField(max_length=32)),
                ("random_field2", models.IntegerField()),
                (
                    "points",
                    django.contrib.gis.db.models.fields.MultiPointField(srid=4326),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="MultiPolygonModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("random_field1", models.CharField(max_length=32)),
                ("random_field2", models.IntegerField()),
                (
                    "polygon",
                    django.contrib.gis.db.models.fields.MultiPolygonField(srid=4326),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="PointModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("random_field1", models.CharField(max_length=32)),
                ("random_field2", models.IntegerField()),
                ("location", django.contrib.gis.db.models.fields.PointField(srid=4326)),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="PolygonModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("random_field1", models.CharField(max_length=32)),
                ("random_field2", models.IntegerField()),
                (
                    "polygon",
                    django.contrib.gis.db.models.fields.PolygonField(srid=4326),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="GeometryModel",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("random_field1", models.CharField(max_length=32)),
                ("random_field2", models.IntegerField()),
                (
                    "points",
                    django.contrib.gis.db.models.fields.GeometryField(srid=4326),
                ),
            ],
            options={"abstract": False},
        ),
    ]



from django.contrib import admin
from django.contrib.gis.admin import ModelAdmin as GeoModelAdmin

from .models import Location


class LocationAdmin(GeoModelAdmin):
    list_display = ('name', 'geometry')


admin.site.register(Location, LocationAdmin)



from django.contrib import admin
from django.contrib.staticfiles import views
from django.urls import include, path

urlpatterns = [
    # Uncomment the next line to enable the admin:
    path('admin/', admin.site.urls),
    # Uncomment the admin/doc line below to enable admin documentation:
    # url('admin/doc/', include('django.contrib.admindocs.urls')),
    path('', include('django_restframework_gis_tests.urls')),
    path('static/<path>', views.serve),
]



#!/usr/bin/env python
import os
import sys

from setuptools import find_packages, setup

from rest_framework_gis import get_version

if sys.argv[-1] == 'publish':
    os.system("python setup.py sdist bdist_wheel")
    os.system("twine upload -s dist/*")
    os.system("rm -rf dist build")
    args = {'version': get_version()}
    print("You probably want to also tag the version now:")
    print("  git tag -a %(version)s -m 'version %(version)s'" % args)
    print("  git push --tags")
    sys.exit()

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='djangorestframework-gis',
    version=get_version(),
    license='BSD',
    author='Douglas Meehan',
    author_email='django-rest-framework-gis@googlegroups.com',
    description='Geographic add-ons for Django Rest Framework',
    long_description=long_description,
    url='https://github.com/openwisp/django-rest-framework-gis',
    download_url='https://github.com/openwisp/django-rest-framework-gis/releases',
    platforms=['Platform Indipendent'],
    keywords=['django', 'rest-framework', 'gis', 'geojson'],
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=['djangorestframework'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Topic :: Internet :: WWW/HTTP',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Framework :: Django',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
    ],
    project_urls={
        'Bug Reports': 'https://github.com/openwisp/django-rest-framework-gis/issues',
        'Continuous Integration': 'https://travis-ci.org/openwisp/django-rest-framework-gis',
        'Mailing List': 'https://groups.google.com/forum/#!forum/django-rest-framework-gis',
        'Code Coverage': 'https://coveralls.io/github/openwisp/django-rest-framework-gis',
        'Source Code': 'https://github.com/openwisp/django-rest-framework-gis',
    },
)



import json
import urllib
from unittest import skipIf

from django.conf import settings
from django.contrib.gis.geos import GEOSGeometry, Polygon
from django.test import TestCase

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse

from .models import Location
from .views import (
    GeojsonLocationContainedInBBoxList,
    GeojsonLocationOrderDistanceToPointList,
    GeojsonLocationWithinDistanceOfPointList,
)

has_spatialite = (
    settings.DATABASES['default']['ENGINE']
    == 'django.contrib.gis.db.backends.spatialite'
)

try:
    from django.contrib.gis.db.models.functions import GeometryDistance

    has_geometry_distance = GeometryDistance and True
except ImportError:
    has_geometry_distance = False


class TestRestFrameworkGisFilters(TestCase):
    """
    unit tests for filters feature in restframework_gis
    """

    def setUp(self):
        self.location_contained_in_bbox_list_url = reverse(
            'api_geojson_location_list_contained_in_bbox_filter'
        )
        self.location_overlaps_bbox_list_url = reverse(
            'api_geojson_location_list_overlaps_bbox_filter'
        )
        self.location_contained_in_tile_list_url = reverse(
            'api_geojson_location_list_contained_in_tile_filter'
        )
        self.location_overlaps_tile_list_url = reverse(
            'api_geojson_location_list_overlaps_tile_filter'
        )
        self.location_within_distance_of_point_list_url = reverse(
            'api_geojson_location_list_within_distance_of_point_filter'
        )
        self.location_within_degrees_of_point_list_url = reverse(
            'api_geojson_location_list_within_degrees_of_point_filter'
        )
        self.geojson_contained_in_geometry = reverse(
            'api_geojson_contained_in_geometry'
        )
        self.location_order_distance_to_point = reverse(
            'api_geojson_location_order_distance_to_point_list_filter'
        )

        treasure_island_geojson = """{
            "type": "Polygon",
            "coordinates": [
                [
                    [
                        -122.44640350341795,
                        37.86103094116189
                    ],
                    [
                        -122.44262695312501,
                        37.85506751416839
                    ],
                    [
                        -122.43481636047363,
                        37.853305500228025
                    ],
                    [
                        -122.42975234985352,
                        37.854660899304704
                    ],
                    [
                        -122.41953849792479,
                        37.852627791344894
                    ],
                    [
                        -122.41807937622069,
                        37.853305500228025
                    ],
                    [
                        -122.41868019104004,
                        37.86211514878027
                    ],
                    [
                        -122.42391586303711,
                        37.870584971740065
                    ],
                    [
                        -122.43035316467285,
                        37.8723465726078
                    ],
                    [
                        -122.43515968322752,
                        37.86963639998042
                    ],
                    [
                        -122.43953704833984,
                        37.86882332875222
                    ],
                    [
                        -122.44640350341795,
                        37.86103094116189
                    ]
                ]
            ]
        }"""
        self.treasure_island_geom = GEOSGeometry(treasure_island_geojson)

        ggpark_geojson = """{
            "type": "Polygon",
            "coordinates": [
                [
                    [
                        -122.5111198425293,
                        37.77125750792944
                    ],
                    [
                        -122.51026153564452,
                        37.76447260365713
                    ],
                    [
                        -122.45309829711913,
                        37.76677954095475
                    ],
                    [
                        -122.45481491088867,
                        37.77424266859531
                    ],
                    [
                        -122.5111198425293,
                        37.77125750792944
                    ]
                ]
            ]
        }"""
        self.ggpark_geom = GEOSGeometry(ggpark_geojson)

    def test_inBBOXFilter_filtering(self):
        """
        Checks that the inBBOXFilter returns only objects strictly contained
        in the bounding box given by the in_bbox URL parameter
        """
        self.assertEqual(Location.objects.count(), 0)

        # Bounding box
        xmin = 0
        ymin = 0
        xmax = 10
        ymax = 10

        url_params = '?in_bbox=%d,%d,%d,%d&format=json' % (xmin, ymin, xmax, ymax)

        # Square with bottom left at (1,1), top right at (9,9)
        isContained = Location()
        isContained.name = 'isContained'
        isContained.geometry = Polygon(((1, 1), (9, 1), (9, 9), (1, 9), (1, 1)))
        isContained.save()

        isEqualToBounds = Location()
        isEqualToBounds.name = 'isEqualToBounds'
        isEqualToBounds.geometry = Polygon(((0, 0), (10, 0), (10, 10), (0, 10), (0, 0)))
        isEqualToBounds.save()

        # Rectangle with bottom left at (-1,1), top right at (5,5)
        overlaps = Location()
        overlaps.name = 'overlaps'
        overlaps.geometry = Polygon(((-1, 1), (5, 1), (5, 5), (-1, 5), (-1, 1)))
        overlaps.save()

        # Rectangle with bottom left at (-3,-3), top right at (-1,2)
        nonIntersecting = Location()
        nonIntersecting.name = 'nonIntersecting'
        nonIntersecting.geometry = Polygon(
            ((-3, -3), (-1, -3), (-1, 2), (-3, 2), (-3, -3))
        )
        nonIntersecting.save()

        # Make sure we only get back the ones strictly contained in the bounding box
        response = self.client.get(
            self.location_contained_in_bbox_list_url + url_params
        )
        self.assertEqual(len(response.data['features']), 2)
        for result in response.data['features']:
            self.assertEqual(
                result['properties']['name'] in ('isContained', 'isEqualToBounds'), True
            )

        # Make sure we get overlapping results for the view which allows bounding box overlaps.
        response = self.client.get(self.location_overlaps_bbox_list_url + url_params)
        self.assertEqual(len(response.data['features']), 3)
        for result in response.data['features']:
            self.assertEqual(
                result['properties']['name']
                in ('isContained', 'isEqualToBounds', 'overlaps'),
                True,
            )

    @skipIf(has_spatialite, 'Skipped test for spatialite backend: not accurate enough')
    def test_TileFilter_filtering(self):
        """
        Checks that the TMSTileFilter returns only objects strictly contained
        in the bounding box given by the tile URL parameter
        """
        self.assertEqual(Location.objects.count(), 0)

        # Bounding box
        z = 1
        x = 1
        y = 0

        url_params = '?tile=%d/%d/%d&format=json' % (z, x, y)

        # Square with bottom left at (1,1), top right at (9,9)
        isContained = Location()
        isContained.name = 'isContained'
        isContained.geometry = Polygon(((1, 1), (9, 1), (9, 9), (1, 9), (1, 1)))
        isContained.save()

        isEqualToBounds = Location()
        isEqualToBounds.name = 'isEqualToBounds'
        isEqualToBounds.geometry = Polygon(
            ((0, 0), (0, 85.05113), (180, 85.05113), (180, 0), (0, 0))
        )
        isEqualToBounds.save()

        # Rectangle with bottom left at (-1,1), top right at (5,5)
        overlaps = Location()
        overlaps.name = 'overlaps'
        overlaps.geometry = Polygon(((-1, 1), (5, 1), (5, 5), (-1, 5), (-1, 1)))
        overlaps.save()

        # Rectangle with bottom left at (-3,-3), top right at (-1,2)
        nonIntersecting = Location()
        nonIntersecting.name = 'nonIntersecting'
        nonIntersecting.geometry = Polygon(
            ((-3, -3), (-1, -3), (-1, 2), (-3, 2), (-3, -3))
        )
        nonIntersecting.save()

        # Make sure we only get back the ones strictly contained in the bounding box
        response = self.client.get(
            self.location_contained_in_tile_list_url + url_params
        )
        self.assertEqual(len(response.data['features']), 2)
        for result in response.data['features']:
            self.assertEqual(
                result['properties']['name'] in ('isContained', 'isEqualToBounds'), True
            )

        # Make sure we get overlapping results for the view which allows bounding box overlaps.
        response = self.client.get(self.location_overlaps_tile_list_url + url_params)
        self.assertEqual(len(response.data['features']), 3)
        for result in response.data['features']:
            self.assertEqual(
                result['properties']['name']
                in ('isContained', 'isEqualToBounds', 'overlaps'),
                True,
            )

    @skipIf(
        has_spatialite, 'Skipped test for spatialite backend: missing feature "dwithin"'
    )
    def test_DistanceToPointFilter_filtering(self):
        """
        Checks that the DistanceFilter returns only objects within the given distance of the
        given geometry defined by the URL parameters
        """
        self.assertEqual(Location.objects.count(), 0)

        # Filter parameters
        distance = 5000  # meters
        point_on_alcatraz = [-122.4222, 37.82667]

        url_params = '?dist=%0.4f&point=hello&format=json' % (distance,)
        response = self.client.get(
            '%s%s' % (self.location_within_distance_of_point_list_url, url_params)
        )
        self.assertEqual(response.status_code, 400)

        url_params = '?dist=%0.4f&point=%0.4f,%0.4f&format=json' % (
            distance,
            point_on_alcatraz[0],
            point_on_alcatraz[1],
        )

        treasure_island = Location()
        treasure_island.name = "Treasure Island"
        treasure_island.geometry = self.treasure_island_geom
        treasure_island.full_clean()
        treasure_island.save()

        ggpark = Location()
        ggpark.name = "Golden Gate Park"
        ggpark.geometry = self.ggpark_geom
        ggpark.save()

        # Make sure we only get back the ones within the distance
        response = self.client.get(
            '%s%s' % (self.location_within_distance_of_point_list_url, url_params)
        )
        self.assertEqual(len(response.data['features']), 1)
        for result in response.data['features']:
            self.assertEqual(result['properties']['name'], treasure_island.name)

        # Make sure we get back all the ones within the distance
        distance = 7000
        url_params = '?dist=%0.4f&point=%0.4f,%0.4f&format=json' % (
            distance,
            point_on_alcatraz[0],
            point_on_alcatraz[1],
        )
        response = self.client.get(
            '%s%s' % (self.location_within_distance_of_point_list_url, url_params)
        )
        self.assertEqual(len(response.data['features']), 2)
        for result in response.data['features']:
            self.assertIn(
                result['properties']['name'], (ggpark.name, treasure_island.name)
            )

        # Make sure we only get back the ones within the distance
        degrees = 0.05
        url_params = '?dist=%0.4f&point=%0.4f,%0.4f&format=json' % (
            degrees,
            point_on_alcatraz[0],
            point_on_alcatraz[1],
        )
        response = self.client.get(
            self.location_within_degrees_of_point_list_url + url_params
        )
        self.assertEqual(len(response.data['features']), 1)
        for result in response.data['features']:
            self.assertEqual(result['properties']['name'], treasure_island.name)

    @skipIf(
        has_spatialite,
        'Skipped test for spatialite backend: missing feature "GeometryDistance"',
    )
    @skipIf(
        not has_geometry_distance,
        'Skipped test for Django < 3.0: missing feature "GeometryDistance"',
    )
    def test_DistanceToPointOrderingFilter_filtering(self):
        """
        Checks that the DistanceOrderingFilter returns the objects in the correct order
        given the geometry defined by the URL parameters
        """
        self.assertEqual(Location.objects.count(), 0)

        url_params = '?point=hello&format=json'
        response = self.client.get(
            '%s%s' % (self.location_order_distance_to_point, url_params)
        )
        self.assertEqual(response.status_code, 400)

        Location.objects.create(
            name='Houston', geometry='SRID=4326;POINT (-95.363151 29.763374)'
        )
        Location.objects.create(
            name='Dallas', geometry='SRID=4326;POINT (-96.801611 32.782057)'
        )
        Location.objects.create(
            name='Oklahoma City', geometry='SRID=4326;POINT (-97.521157 34.464642)'
        )
        Location.objects.create(
            name='Wellington', geometry='SRID=4326;POINT (174.783117 -41.315268)'
        )
        Location.objects.create(
            name='Pueblo', geometry='SRID=4326;POINT (-104.609252 38.255001)'
        )
        Location.objects.create(
            name='Lawrence', geometry='SRID=4326;POINT (-95.235060 38.971823)'
        )
        Location.objects.create(
            name='Chicago', geometry='SRID=4326;POINT (-87.650175 41.850385)'
        )
        Location.objects.create(
            name='Victoria', geometry='SRID=4326;POINT (-123.305196 48.462611)'
        )

        point = [-90, 40]

        url_params = '?point=%i,%i&format=json' % (point[0], point[1])
        response = self.client.get(
            '%s%s' % (self.location_order_distance_to_point, url_params)
        )
        self.assertEqual(len(response.data['features']), 8)
        self.assertEqual(
            [city['properties']['name'] for city in response.data['features']],
            [
                'Chicago',
                'Lawrence',
                'Oklahoma City',
                'Dallas',
                'Houston',
                'Pueblo',
                'Victoria',
                'Wellington',
            ],
        )

        url_params = '?point=%i,%i&order=desc&format=json' % (point[0], point[1])
        response = self.client.get(
            '%s%s' % (self.location_order_distance_to_point, url_params)
        )
        self.assertEqual(len(response.data['features']), 8)
        self.assertEqual(
            [city['properties']['name'] for city in response.data['features']],
            [
                'Wellington',
                'Victoria',
                'Pueblo',
                'Houston',
                'Dallas',
                'Oklahoma City',
                'Lawrence',
                'Chicago',
            ],
        )

    @skipIf(
        has_spatialite,
        'Skipped test for spatialite backend: missing feature "contains_properly"',
    )
    def test_GeometryField_filtering(self):
        """Checks that the GeometryField allows sane filtering."""
        self.assertEqual(Location.objects.count(), 0)

        treasure_island = Location()
        treasure_island.name = "Treasure Island"
        treasure_island.geometry = self.treasure_island_geom
        treasure_island.full_clean()
        treasure_island.save()

        ggpark = Location()
        ggpark.name = "Golden Gate Park"
        ggpark.geometry = self.ggpark_geom
        ggpark.save()

        point_inside_ggpark_geojson = """
        { "type": "Point", "coordinates": [ -122.49034881591797, 37.76949349270407 ] }
        """

        try:
            quoted_param = urllib.quote(point_inside_ggpark_geojson)
        except AttributeError:
            quoted_param = urllib.parse.quote(point_inside_ggpark_geojson)

        url_params = "?contains_properly=%s" % (quoted_param,)

        response = self.client.get(
            '{0}{1}'.format(self.geojson_contained_in_geometry, url_params)
        )
        self.assertEqual(len(response.data), 1)

        geometry_response = GEOSGeometry(json.dumps(response.data[0]['geometry']))

        self.assertTrue(geometry_response.equals_exact(self.ggpark_geom))
        self.assertEqual(response.data[0]['name'], ggpark.name)

        # try without any param, should return both
        response = self.client.get(self.geojson_contained_in_geometry)
        self.assertEqual(len(response.data), 2)

    def test_inBBOXFilter_filtering_none(self):
        url_params = '?in_bbox=&format=json'
        response = self.client.get(
            self.location_contained_in_bbox_list_url + url_params
        )
        self.assertDictEqual(
            response.data, {'type': 'FeatureCollection', 'features': []}
        )

    def test_inBBOXFilter_ValueError(self):
        url_params = '?in_bbox=0&format=json'
        response = self.client.get(
            self.location_contained_in_bbox_list_url + url_params
        )
        self.assertEqual(
            response.data['detail'],
            'Invalid bbox string supplied for parameter in_bbox',
        )

    def test_inBBOXFilter_filter_field_none(self):
        original_value = GeojsonLocationContainedInBBoxList.bbox_filter_field
        GeojsonLocationContainedInBBoxList.bbox_filter_field = None
        url_params = '?in_bbox=0,0,0,0&format=json'
        response = self.client.get(
            self.location_contained_in_bbox_list_url + url_params
        )
        self.assertDictEqual(
            response.data, {'type': 'FeatureCollection', 'features': []}
        )
        GeojsonLocationContainedInBBoxList.bbox_filter_field = original_value

    def test_TileFilter_filtering_none(self):
        url_params = '?tile=&format=json'
        response = self.client.get(
            self.location_contained_in_tile_list_url + url_params
        )
        self.assertEqual(response.data, {'type': 'FeatureCollection', 'features': []})

    def test_TileFilter_ValueError(self):
        url_params = '?tile=1/0&format=json'
        response = self.client.get(
            self.location_contained_in_tile_list_url + url_params
        )
        self.assertEqual(
            response.data['detail'], 'Invalid tile string supplied for parameter tile'
        )

    def test_DistanceToPointFilter_filtering_none(self):
        url_params = '?dist=5000&point=&format=json'
        response = self.client.get(
            '%s%s' % (self.location_within_distance_of_point_list_url, url_params)
        )
        self.assertDictEqual(
            response.data, {'type': 'FeatureCollection', 'features': []}
        )

    def test_DistanceToPointFilter_filter_field_none(self):
        original_value = GeojsonLocationWithinDistanceOfPointList.distance_filter_field
        GeojsonLocationWithinDistanceOfPointList.distance_filter_field = None
        url_params = '?dist=5000&point=&format=json'
        response = self.client.get(
            '%s%s' % (self.location_within_distance_of_point_list_url, url_params)
        )
        self.assertDictEqual(
            response.data, {'type': 'FeatureCollection', 'features': []}
        )
        GeojsonLocationWithinDistanceOfPointList.distance_filter_field = original_value

    def test_DistanceToPointFilter_ValueError_point(self):
        url_params = '?dist=500.0&point=hello&format=json'
        response = self.client.get(
            '%s%s' % (self.location_within_distance_of_point_list_url, url_params)
        )
        self.assertEqual(
            response.data['detail'],
            'Invalid geometry string supplied for parameter point',
        )

    def test_DistanceToPointFilter_ValueError_distance(self):
        url_params = '?dist=wrong&point=12.0,42.0&format=json'
        response = self.client.get(
            '%s%s' % (self.location_within_distance_of_point_list_url, url_params)
        )
        self.assertEqual(
            response.data['detail'],
            'Invalid distance string supplied for parameter dist',
        )

    @skipIf(
        not has_geometry_distance,
        'Skipped test for Django < 3.0: missing feature "GeometryDistance"',
    )
    def test_DistanceToPointOrderingFilter_filtering_none(self):
        url_params = '?point=&format=json'
        response = self.client.get(
            '%s%s' % (self.location_order_distance_to_point, url_params)
        )
        self.assertDictEqual(
            response.data, {'type': 'FeatureCollection', 'features': []}
        )

    @skipIf(
        not has_geometry_distance,
        'Skipped test for Django < 3.0: missing feature "GeometryDistance"',
    )
    def test_DistanceToPointOrderingFilter_ordering_filter_field_none(self):
        original_value = (
            GeojsonLocationOrderDistanceToPointList.distance_ordering_filter_field
        )
        GeojsonLocationOrderDistanceToPointList.distance_ordering_filter_field = None
        url_params = '?point=&format=json'
        response = self.client.get(
            '%s%s' % (self.location_order_distance_to_point, url_params)
        )
        self.assertDictEqual(
            response.data, {'type': 'FeatureCollection', 'features': []}
        )
        GeojsonLocationOrderDistanceToPointList.distance_ordering_filter_field = (
            original_value
        )

    @skipIf(has_geometry_distance, 'Skipped test for Django >= 3.0')
    def test_DistanceToPointOrderingFilter_not_available(self):
        url_params = '?point=12,42&format=json'
        with self.assertRaises(ValueError):
            self.client.get(
                '%s%s' % (self.location_order_distance_to_point, url_params)
            )
